//Complex interval Newton method to compute a verified enclosure of
//a zero of a complex function f(z).
//We use automatic differentiation to compute the derivative of f(z). 

#include <iostream>    
using namespace std;

#include <cinterval.hpp>  // complex intervals
#include "citaylor.hpp"   // automatic differentiation is used 
                          // for evaluating f'(z)

using namespace cxsc;           
using namespace taylor;  // automatic differentiation

// definition of f for computing Taylor coefficients
citaylor f(citaylor z) {
   const cinterval c(interval(8),interval(1));  // c:=8+i
   const cinterval a(interval(0),interval(4));  // a:=4*i
   return atan( (z-a)*ln(sqr(z)-5*z+c) );
}   

// inclusion function for f via citaylor
cinterval f(const cinterval &w)  
{ 
   citaylor z(0,w); // Taylor coefficients up to order 0
   return get_j_derivative(f(z), 0);  
} 

// inclusion function for f' via citaylor, using automatic differentiation
cinterval df(const cinterval &w) 
{ 
   citaylor z(1,w); // Taylor coefficients up to order 1
   return get_j_derivative(f(z), 1); 
} 

// main procedure: interval Newton method 
int main()
{            
  cout << boolalpha << SetPrecision(20,15);   
  cinterval m, dfxk, Nxk, xkp1; 
  bool verified = false; 
  
  // starting interval
  cinterval xk = cinterval(interval(1.997,2.007),interval(0.996,1.0091));
  cout << "Iterate " << 0 << ": " << xk << endl; 
  
  // iteration
  for (int k=1; k<5; k++) { 
     m = mid(xk);           // midpoint
     dfxk = df(xk);         // enclosure of f'
     Nxk  = m - f(m)/dfxk;  // interval Newton operator
     xkp1 = Nxk & xk;       // next iterate   
     cout << "Iterate " << k << ": " << xkp1 << endl;

     verified |= (Nxk < xk);  //check for proper inclusion
     cout << "Verification successful: " << verified << endl;
     
     if (xk == xkp1) break; // no further improvement possible

     xk = xkp1;             // replace old iterate by new iterate, continue
  }       
  return 0;
}

/* Generated output:
Iterate 0: ([   1.997000000000000,   2.007000000000001],
            [   0.995999999999999,   1.009100000000001])
Iterate 1: ([   1.997000000000000,   2.001686857904093],
            [   0.998244050366883,   1.001446655588546])
Verification successful: false
Iterate 2: ([   1.999935121034121,   2.000075438075472],
            [   0.999931959857042,   1.000070155357838])
Verification successful: true
Iterate 3: ([   1.999999987372335,   2.000000012633344],
            [   0.999999987334793,   1.000000012602762])
Verification successful: true
Iterate 4: ([   1.999999999999999,   2.000000000000001],
            [   0.999999999999999,   1.000000000000001])
Verification successful: true
*/
